@@ -7,7 +7,7 @@
 
 import logging
 import math
-from typing import MutableMapping, Union, Sequence
+from typing import Iterable, MutableMapping, Union, Sequence
 
 import numpy as np
 
@@ -54,8 +54,8 @@ class DiscreteSampling():
                 'required_args': ['_names', '_ranges', '_numberOfSamples', '_includeBoundingBox']
             },
             'normalLhs': {
-                'required_args':
-                ['_names', '_ranges', '_numberOfSamples', '_mu', '_sigma', '_includeBoundingBox']
+                'required_args': ['_names', '_numberOfSamples', '_mu', '_sigma'],
+                'optional_args': ['_ranges', '_cov']
             },
             'randNormal': {
                 'required_args': [
@@ -110,6 +110,10 @@ class DiscreteSampling():
         # determine the dimension (=number of fields)
         self.number_of_fields = len(self.fields)
 
+        # extra bounding box samples are not treated by lhs algorithm, however part of the lists
+        self.number_of_samples = 0
+        self.number_of_bb_samples = 0
+
     def generate(self) -> dict:
         '''
         '''
@@ -191,47 +195,105 @@ class DiscreteSampling():
                     }
                 )
 
+        if self.sampling_type == 'normalLhs':
+            '''lhs using gaussian normal dstributions
+            required input arguments:
+            * _names: required names template
+            * _numberOfSamples: required how many samples
+            * _mu: required absolute location vector of distribution center point (origin)
+            * _sigma: variation (vector), or required scalar, optional vector, optional cov
+            or
+            * _cov: @ _mu optional rotation (tensor), otherwise I(_numberOfSamples,_numberOfSamples)
+            '''
+            self.number_of_samples = int(self.kwargs['_numberOfSamples'])
+            self.leading_zeros = int(math.log10(self.number_of_samples) - 1.e-06) + 1
+            self.case_names = [
+                '%s_%s' % (self.base_name, format(i, '0%i' % self.leading_zeros))
+                for i in range(self.number_of_samples)
+            ]
+            return_dict.update({'_case_name': self.case_names})
+
+            # check dimension of nested lists / vectors
+            if not len(self.kwargs['_names']) == len(self.kwargs['_mu']):
+                logger.error('lists _names and _mu: lenght of entries do not match')
+            self.mean = self.kwargs['_mu']
+            if isinstance(self.kwargs['_sigma'],
+                          list) and not len(self.kwargs['_names']) == len(self.kwargs['_sigma']):
+                logger.error('lists _names and _sigma: lenght of entries do not match')
+            self.std = self.kwargs['_sigma']
+
+            self.variables = self.generate_normal_lhs()
+
+            # optional clipping cube
+            # ToDo:
+            # clipping is implemented, resets the value to range value
+            # if this is unaffordable, purgign should be considered, requiring a decrementation of number_of_samples
+            if '_ranges' in self.kwargs:
+                # check equality of nested lists
+                for item in self.kwargs['_ranges']:
+                    if not len(item) == 2:
+                        logger.error('ranges: lists in list of lists do not contain min and max')
+
+                self.bounds = self.kwargs['_ranges']
+
+                clipped = np.zeros(self.variables.shape)
+
+                for index, field in enumerate(self.variables):
+                    clipped[index] = np.clip(field, self.bounds[index][0], self.bounds[index][1])
+
+                self.variables = clipped
+
+            for index, item in enumerate(self.fields):
+                return_dict.update({self.fields[index]: self.variables[index].tolist()})
+
         if self.sampling_type == 'uniformLhs':
             '''lhs uniform
             later implementations may change thie section
             '''
+            # first dimension n samples
             self.number_of_samples = int(self.kwargs['_numberOfSamples'])
+            if self.kwargs['_includeBoundingBox'] is True:
+                self.number_of_bb_samples = int(2**self.number_of_fields)
+            self.number_of_samples += self.number_of_bb_samples
+
             self.leading_zeros = int(math.log10(self.number_of_samples) - 1.e-06) + 1
             self.bounds = self.kwargs['_ranges']
-
-            if self.kwargs['_includeBoundingBox'] is True:
-                self.number_of_samples += int(2**self.number_of_fields)
+            self.case_names = [
+                '%s_%s' % (self.base_name, format(i, '0%i' % self.leading_zeros))
+                for i in range(self.number_of_samples)
+            ]
+            return_dict.update({'_case_name': self.case_names})
 
             # check length of nested lists
             if not len(self.kwargs['_names']) == len(self.kwargs['_ranges']):
-                logger.error('lists: lenght of entries do not match')
+                logger.error('lists _names and _ranges: lenght of entries do not match')
 
             # check equality of nested lists
             for item in self.kwargs['_ranges']:
                 if not len(item) == 2:
                     logger.error('ranges: lists in list of lists do not contain min and max')
 
-            self.variables = self.generate_lhs()
-
-            # output
-            return_dict.update(
-                {
-                    '_case_name': [
-                        '%s_%s' % (self.base_name, format(i, '0%i' % self.leading_zeros))
-                        for i in range(self.number_of_samples)
-                    ]
-                }
-            )
+            self.variables = self.generate_uniform_lhs()
 
             if self.kwargs['_includeBoundingBox'] is True:
-                # permutate boundaries
-                tmp = list(itertools.product(self.kwargs['_ranges'][0], self.kwargs['_ranges'][1]))
-                for field_index in range(1, self.number_of_fields - 1):
-                    tmp = list(itertools.product(tmp, self.kwargs['_ranges'][field_index + 1]))
+                tmp = []
+                if len(self.kwargs['_ranges']) == 1:
+                    # (only one single parameter defined)
+                    tmp = list(self.kwargs['_ranges'][0])
+                else:
+                    # permutate boundaries of all paramaters
+                    tmp = list(
+                        itertools.product(self.kwargs['_ranges'][0], self.kwargs['_ranges'][1])
+                    )
+                    for field_index in range(1, self.number_of_fields - 1):
+                        tmp = list(itertools.product(tmp, self.kwargs['_ranges'][field_index + 1]))
 
                 self.boundingBox = []
                 for item in tmp:
-                    self.boundingBox.append(list(self.flatten(item)))
+                    if isinstance(item, Iterable):
+                        self.boundingBox.append(list(self.flatten(item)))
+                    else:
+                        self.boundingBox.append([item])
 
                 self.variables = np.concatenate(
                     (np.array(self.boundingBox).T, self.variables), axis=1
@@ -299,14 +361,20 @@ class DiscreteSampling():
         if self.sampling_type == 'sobol':
             '''
             '''
-
+            # first dimension n samples
             self.number_of_samples = int(self.kwargs['_numberOfSamples'])
+            if self.kwargs['_includeBoundingBox'] is True:
+                self.number_of_bb_samples = int(2**self.number_of_fields)
+            self.number_of_samples += self.number_of_bb_samples
+
             self.leading_zeros = int(math.log10(self.number_of_samples) - 1.e-06) + 1
             self.onset = int(self.kwargs['_onset'])
             self.bounds = self.kwargs['_ranges']
-
-            if self.kwargs['_includeBoundingBox'] is True:
-                self.number_of_samples += int(2**self.number_of_fields)
+            self.case_names = [
+                '%s_%s' % (self.base_name, format(i, '0%i' % self.leading_zeros))
+                for i in range(self.number_of_samples)
+            ]
+            return_dict.update({'_case_name': self.case_names})
 
             # check length of nested lists
             if not len(self.kwargs['_names']) == len(self.kwargs['_ranges']):
@@ -319,15 +387,6 @@ class DiscreteSampling():
 
             self.variables = self.generate_sobol()
 
-            return_dict.update(
-                {
-                    '_case_name': [
-                        '%s_%s' % (self.base_name, format(i, '0%i' % self.leading_zeros))
-                        for i in range(self.number_of_samples)
-                    ]
-                }
-            )
-
             if self.kwargs['_includeBoundingBox'] is True:
                 # permutate boundaries
                 tmp = list(itertools.product(self.kwargs['_ranges'][0], self.kwargs['_ranges'][1]))
@@ -364,8 +423,12 @@ class DiscreteSampling():
 
         return scale * field + range[0] - field.min(axis=0) * scale
 
-    def generate_lhs(self):
+    def generate_uniform_lhs(self):
         '''
+        alternative
+        from pyDOE import lhs
+        lhs(n, [samples, criterion, iterations])
+        criterion: center|maximin|centermaximin|correlation
         '''
         from SALib.sample import latin
 
@@ -375,11 +438,57 @@ class DiscreteSampling():
             'bounds': self.bounds,
         }
 
-        return latin.sample(problem, self.number_of_samples).T
+        return latin.sample(problem, self.number_of_samples - self.number_of_bb_samples).T
+
+    # ToDo: implementation of cov (spacial rotation)
+    def generate_normal_lhs(self):
+        '''gaussnormal lhs
+        '''
+        from pyDOE import lhs
+        from scipy.stats import norm    # qmc, truncnorm
+
+        lhs_distribution = lhs(self.number_of_fields, samples=self.number_of_samples - self.number_of_bb_samples, criterion="corr")
+        #criterion:center|c: center the points within the sampling intervals
+        #          maximin|m: maximize the minimum distance between points, but place the point in a randomized location within its interval
+        #          centermaximin|cm: same as “maximin”, but centered within the intervals
+        #          correlation|corr: minimize the maximum correlation coefficient
+        # lhs_distribution = qmc.LatinHypercube(d=self.number_of_fields, optimization="random-cd").random(n=self.number_of_samples - self.number_of_bb_samples)
+
+        # convert to array does a better identification
+        self.std = np.array(self.std)
+
+        if self.std.shape == ():
+            # scalar value given
+            sample_set = norm(loc=self.mean, scale=self.std).ppf(lhs_distribution)
+
+        elif self.std.shape == (self.number_of_fields, ):
+            # list given
+
+            sample_set = norm(loc=self.mean, scale=self.std).ppf(lhs_distribution)
+
+        elif self.std.shape == (self.number_of_fields, self.number_of_fields):
+            # matrix given
+            # todo:
+            # - estimate correct transform
+            # - derive rotation from local cov
+            # - find out how (if) non-uniform scale and rotate can be done in one operation
+            # scale as normal
+            sample_set = norm(loc=self.mean, scale=np.diag(self.std)).ppf(lhs_distribution)
+
+            # rotate
+            #from scipy.spatial.transform import Rotation as R # works only in R3
+            #r = R.from_matrix(self.std)
+            #sample_set = np.array([r.apply(x) for x in sample_set])
+
+        else:
+            logger.error('something went wrong: %s' % str(self.std.shape))
+
+        # transpose to be aligned with uniformLhs output
+        return sample_set.T
 
     # @TODO: Should be reimplemented using the scipy.stats.qmc.sobol
     #        https://scipy.github.io/devdocs/reference/generated/scipy.stats.qmc.Sobol.html#scipy-stats-qmc-sobol
-    #        This to substitute the sobol-seq package, which is no longer aintained.
+    #        This to substitute the sobol-seq package, which is no longer maintained.
     #        (See https://github.com/naught101/sobol_seq)
     #        CLAROS, 2022-05-27
     def generate_sobol(self):
@@ -388,10 +497,11 @@ class DiscreteSampling():
         import sobol_seq
 
         sequence = sobol_seq.i4_sobol_generate(
-            self.number_of_fields, self.number_of_samples + self.onset
+            self.number_of_fields, self.number_of_samples - self.number_of_bb_samples + self.onset
         )
 
-        sample_set = sequence[self.onset:self.onset + self.number_of_samples].T
+        sample_set = sequence[self.onset:self.onset + self.number_of_samples
+                              - self.number_of_bb_samples].T
 
         for index, item in enumerate(sample_set):
             sample_set[index] = self.min_max_scale(item, self.bounds[index])
